/****************************************************************
 *
 * Xproc.cc : Weakest Othello Championship X GUI Main routine.
 *
 * (c) 1997 Yutaka Oiwa (Aleph-0). All rights reserved.
 *
 * This program includes Unix-CUI othello routines.
 * original source is "othello.c" (c) 1997 Gel.
 *
 * $Id: Xproc.cc,v 1.4 1997/10/22 15:05:50 yutaka Exp $	
 *
 ****************************************************************/

#define VERSION "0.4.0"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <time.h>
#include <limits.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xresource.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Viewport.h>
#include <X11/xpm.h>

#include "othello.h"
#include "timeAPI.h"

////////////////////////////////////////////////////////////////
// Othello routines
////////////////////////////////////////////////////////////////

// think_t and turn reffered from timeAPI
static int player_id[2] = {0, 1}; // user routine NO
long think_t[2]; // think time
static bool timeout_p[2];
participant_t *think_fun[2]; // think function
static char *player_name[2] = {"", ""}; // player name
static char m_board[64]; // master board
static char v_board[64]; // visible board
static char board[64]; // board passed to user routine
static int hands; // count hands
int turn; // turn (0-1)
static int last_hand; // last stone position
static int time_limit; // time limit in microseconds

const int direction_x[] = {-1,  0,  1, -1,    1, -1,  0,  1};   /* 0 1 2 */
const int direction_y[] = {-1, -1, -1,  0,    0,  1,  1,  1};   /* 3 @ 4 */
const int direction[] =   {-9, -8, -7, -1,    1,  7,  8,  9};   /* 5 6 7 */

// dummy routine
extern "C" int human(char *, int)
{
  abort();
}

// initialize board
static void InitBoard(char *m_board)
{
  /* Set Initial State */
  for(int i = 0; i < 64; i++)
    m_board[i] = SPACE;

  m_board[27] = m_board[36] = WHITE; /*  */
  m_board[28] = m_board[35] = BLACK; /*  */
}

// initialize (1st stage: internal)
static void initiate_game_1()
{
  think_fun[0] = e_list[player_id[0]].func;
  think_fun[1] = e_list[player_id[1]].func;
  think_t[0] = think_t[1] = 0;
  timeout_p[0] = timeout_p[1] = 0;

  turn = 0;
  hands = 1;

  InitBoard(m_board);
  last_hand = -1;
}

// initialize (2nd stage: user routine)
static void initiate_game_2()
{
  char board[64];
  CopyBoard(m_board, board, NORM);
  if(think_fun[0])
    think_fun[0](board, INIT_AS_BLACK);

  CopyBoard(m_board, board, REVERSE);
  if(think_fun[1])
    think_fun[1](board, INIT_AS_BLACK);
}

// remove any marks put on board
static void RepairBoard(char *board)
{
  int i;

  for(i = 0; i < 64; i++)
    if(board[i] < 0)
      board[i] = -board[i];
}

// endgame-p?
static bool isEndgame(void)
{
  for(int i = 0; i <= 63; i++) {
    if (Put(m_board, i, BLACK, NOT_CHANGE)
	|| Put(m_board, i, WHITE, NOT_CHANGE))
      return false;
  }
  return true;
}

// pass-p?
static bool isPass(char *board, int c)
{
  for(int i = 0; i <= 63; i++) {
    if (Put(board, i, c, NOT_CHANGE))
      return false;
  }
  return true;
}

int internal_kirimomi(char *board, int status)
{
  int i, get_max, p, tmp;

  get_max = 0;
  p = 0;

  if(passp(status) || initp(status))
    return 0;

  for(i = 0; i < 64; i++){
    tmp = Put(board, i, MY_COLOR, NOT_CHANGE);
    if((tmp > 0) && (get_max < tmp)){
      p = i;
      get_max = tmp;
    }
  }

  return p;
}


////////////////////////////////////////////////////////////////
// X routines
////////////////////////////////////////////////////////////////

// Global X toolkit Vars
static Display *X_d = 0;
static Window X_root = 0;
static XtAppContext X_app = NULL;
static Atom XA_wm_protocols, XA_wm_delete_window;

// widgets

static Widget XW_shell, XW_form, XW_board, XW_cell[64], XW_message, XW_handslog;
static Widget XW_scores;
static Widget XW_score_board[2], XW_score_name[2], XW_score_num[2], XW_score_time[2];
static Widget XW_nav_bar, XW_pass_btn;
static Widget XW_command_bar, XW_start_btn, XW_quit_btn, XW_savelog_btn;
static Widget XW_statusbar, XW_turn_num;

// widgets of popup
static Widget XpW_popup, XpW_frame, XpW_f0, XpW_f1, XpW_label0, XpW_label1, XpW_list0, XpW_list1;
static Widget XpW_okbtn, XpW_cancelbtn;

// GUI status

static enum {
  COMPTURN, HUMANTURN, HUMANPASS, THINKING, NOTSTARTED
} X_status = NOTSTARTED;

// application resource
static struct X_appres_t {
  Bool help; // -h
  int interval; // computer wait
  int timelimit; // time limit in millisec
  Pixel gray_color; // gray color used to show

  String logsavefile;

  // UI strings
  String str_Bcaps, str_Wcaps, str_B, str_W;
  String str_timeover, str_none, str_over;
  String str_handslog, str_gameend, str_gamedraw, str_violation;
  String str_compturn, str_humanpass, str_humanturn;
} X_appres;

// others
static Pixmap pix_empty, pix_black[4], pix_white[4];
static XtIntervalId XTO_compwait, XTO_human;
static Pixel XCOL_dgreen, XCOL_gray;
static char *X_Colorname[2];
static char *X_colorname[2];

// log
static time_t X_starttime;
static char X_logstring[128 * 4]; // 4bytes * 120hands upper bound

/////////////////////
// Utility routines
/////////////////////

#define XtRMilliInt "MilliInt"

static Boolean X_Convert_MagInt_internal(XrmValue *from, XrmValue *to, int mag, char *type)
{
  char *ptr;

  double d = strtod((char *)from->addr, &ptr);

  if (ptr == from->addr || *ptr) {
    XtStringConversionWarning(from->addr, type);
    return False;
  }

  int r = (int)(d * mag + 0.5);

  *(int *)to->addr = r;
  return True;
}

static Boolean X_TypeConvert_MilliInt(Display *, XrmValue *, Cardinal *,
				      XrmValue *from, XrmValue *to, XtPointer *) {
  return X_Convert_MagInt_internal(from, to, 1000, XtRMilliInt);
}

inline static void X_SetButtonActive(Widget w, Bool f)
{
  XtVaSetValues(w, XtNsensitive, f, NULL);
}

///////////////////////////////
// Display Widget Manipulates
///////////////////////////////

// Output Message to message bar
static void X_message(char *format, ...)
{
  va_list va;
  va_start(va, format);
  static char buf[256];
  vsprintf(buf, format, va);

  XtVaSetValues(XW_message,
		XtNlabel, buf,
		NULL);
 
  va_end(va);
}

static void X_put_handslog(int hand)
{
  char buf[10], buf2[60];

  if (hand == -1)
    strcpy(buf, "Pass");
  else
    sprintf(buf, "%c%c", (hand % 8) + 'A', hand / 8 + '1');

  sprintf(buf2, X_appres.str_handslog, hands, X_colorname[turn],
	  player_name[turn], buf);

  XtVaSetValues(XW_handslog,
		XtNlabel, buf2,
		NULL);

  sprintf(buf, "%.2d ", hand);

  strcat(X_logstring, buf);
}

// Board keeping
static Pixmap X_currentpic[63];
static bool X_ishilighted[63];

static void X_hilight_cell(int p, bool f)
{
  if (X_ishilighted[p] != f) {
    X_ishilighted[p] = f;
    XtVaSetValues(XW_cell[p],
		  XtNbackground, (f ? XCOL_gray : XCOL_dgreen),
		  NULL);
  }
}

static void X_unhilight_board()
{
  for (int i = 0; i <= 63; i++) {
    X_hilight_cell(i, false);
  }
}

static void X_setboard(char *board, int lasthand)
{
  for(int i = 0; i <= 63; i++) {
    int p = board[i];
    Pixmap New;
    int o = (i == lasthand) ? 3 : (p < 0 ? 2 : 0);

    if (p == 0) 
      New = pix_empty;
    else if (p == BLACK || p == -BLACK)
      New = pix_black[o];
    else
      New = pix_white[o];

    if (X_currentpic[i] != New) {
      X_currentpic[i] = New;
      XtVaSetValues(XW_cell[i], XtNbitmap, New, NULL);
    }
  }
}

void convert_timestring(char *buf, int t, int f)
{
  int s = t / 1000000L;
  long us = (t % 1000000L);
  if (s >= 3600)
    sprintf(buf, "%d:%02d:%02d", s / 3600, s / 3600 % 60, s % 60);
  else if (s >= 600 || (f && s > 60))
    sprintf(buf, "%2d:%02d.%1ld", s / 60, s % 60, us / 100000);
  else if (s >= 60)
    sprintf(buf, "%1d:%02d.%02ld", s / 60, s % 60, us / 10000);
  else if (f)
    sprintf(buf, "%2d.%1ld", s % 60, us / 100000);
  else if (s >= 10)
    sprintf(buf, "%2d.%03ld", s, us / 1000);
  else
    sprintf(buf, "%1d.%04ld", s, us / 100);
}

// timer box
static void X_update_time(int i, long t, bool timeout, int f)
{
  char buf[32];

  if (timeout) {
    XtVaSetValues(XW_score_time[i], XtNlabel, X_appres.str_over, NULL);
    return;
  }

  convert_timestring(buf, t, f);
  XtVaSetValues(XW_score_time[i], 
		XtNlabel, buf,
		NULL);
}  

// update status displays
static void X_update_state()
{
  char buf[32];

  X_setboard(v_board, last_hand);
  sprintf(buf, "%d", hands);
  XtVaSetValues(XW_turn_num, XtNlabel, buf, NULL);

  int b, w;

  CountDisks(m_board, &b, &w);

  sprintf(buf, "%d", b);
  XtVaSetValues(XW_score_num[0], XtNlabel, buf, NULL);
  sprintf(buf, "%d", w);
  XtVaSetValues(XW_score_num[1], XtNlabel, buf, NULL);

  for (int i = 0; i < 2; i++) {
    X_update_time(i, think_t[i], timeout_p[i], false);
  }
}

/////////////////////////
// Othello GUI Routines
/////////////////////////

// Change Widget status for waiting start

static void X_setstate_startwait()
{
  X_status = NOTSTARTED;

  for(int i = 0; i <= 63; i++) {
    X_SetButtonActive(XW_cell[i], False);
  }
  X_SetButtonActive(XW_pass_btn, False);
  X_SetButtonActive(XW_start_btn, True);

  if (XTO_human) {
    XtRemoveTimeOut(XTO_human);
    XTO_human = 0;
  }
  if (XTO_compwait) {
    XtRemoveTimeOut(XTO_compwait);
    XTO_compwait = 0;
  }
}

static void X_setup_for_next();

// put stone and go next turn

static void X_put_stone(int i)
{
  Put(m_board, i, turn + 1, R_CHANGE);
  CopyBoard(m_board, v_board, NORM);
  RepairBoard(m_board);
  X_put_handslog(i);

  last_hand = i;
  turn ^= 1;

  X_update_state();
  
  X_setup_for_next();
}

// pass and go next turn

static void X_pass_stone()
{
  X_put_handslog(-1);
  turn ^= 1;
  last_hand = -1;
  
  RepairBoard(v_board);
  
  X_update_state();
  X_setup_for_next();
}

static void XTE_comp_timer_event(XtPointer cdat, XtIntervalId *);

static void X_computer_timeover()
{
  think_fun[turn] = internal_kirimomi;
  timeout_p[turn] = true;

  X_update_state();
  X_message(X_appres.str_timeover,
	    X_Colorname[turn], player_name[turn]);

  // rescheduling

  XTO_compwait = XtAppAddTimeOut(X_app, X_appres.interval, XTE_comp_timer_event, 
				 (XtPointer)(hands * 2 + turn));
}

// Timer Event: start thinking

static void XTE_comp_timer_event(XtPointer cdat, XtIntervalId *)
{
  XTO_compwait = 0; // automatically removed by Xt

  if (X_status != COMPTURN
      || (int)cdat != hands * 2 + turn)
    return;

  CopyBoard(m_board, board, (turn ? REVERSE : NORM));

  if (isPass(board, MY_COLOR)) {
    SetTime();
    think_fun[turn](board, PASS);
    StopTime();
    if (!timeout_p[turn] && think_t[turn] > time_limit) {
      X_computer_timeover();
    } else {
      X_pass_stone();
    }
  } else {
    SetTime();
    int p = think_fun[turn](board, NORM);
    StopTime();
  
    if (p < 0 || p > 63 || Put(board, p, MY_COLOR, NOT_CHANGE) < 1) {
      X_update_state();
      XtVaSetValues(XW_score_num[turn], 
		    XtNlabel, "64", NULL);
      XtVaSetValues(XW_score_num[1-turn], 
		    XtNlabel, "0", NULL);
      X_message(X_appres.str_violation, X_Colorname[turn],
		player_name[turn], p);
      X_setstate_startwait();
    } else
      if (!timeout_p[turn] && think_t[turn] > time_limit)
	X_computer_timeover();
      else
	X_put_stone(p);
  }
}

static void X_start_human_update_timer();

// Timer Event: update human's stopwatch

static void XTE_human_timer_event(XtPointer cdat, XtIntervalId *)
{
  if ((X_status != HUMANTURN && X_status != HUMANPASS)
      || (int)cdat != hands * 2 + turn)
    return;

  X_update_time(turn, GetTime(), 0, true);

  X_start_human_update_timer();
}

static void X_start_human_update_timer()
{
  XTO_human = XtAppAddTimeOut(X_app, 100, XTE_human_timer_event, 
			      (XtPointer)(hands * 2 + turn));
}

static void X_stop_human_update_timer()
{
  XtRemoveTimeOut( XTO_human );
  XTO_human = 0;
}

// setup for next turn

static void X_setup_for_next()
{
  if (isEndgame()) {
    X_update_state();

    int b, w;
    CountDisks(m_board, &b, &w);

    if (b > w)
      X_message(X_appres.str_gameend, X_appres.str_Wcaps, player_name[1], b - w);
    else if (b < w)
      X_message(X_appres.str_gameend, X_appres.str_Bcaps, player_name[0], w - b);
    else
      X_message(X_appres.str_gamedraw);

    X_SetButtonActive(XW_savelog_btn, True);

    X_setstate_startwait();
    return;
  }
  if (last_hand >= 0)
    hands++;

  X_update_state();
  
  if (think_fun[turn]) {
    X_status = COMPTURN;
    X_message(X_appres.str_compturn, X_Colorname[turn], player_name[turn]);
    for(int i = 0; i <= 63; i++) {
      X_SetButtonActive(XW_cell[i], False);
    }
    X_SetButtonActive(XW_pass_btn, False);

    XTO_compwait = XtAppAddTimeOut(X_app, X_appres.interval, XTE_comp_timer_event, 
				   (XtPointer)(hands * 2 + turn));
  } else {
    X_status = HUMANTURN;
    // Human
    
    int passp = 1;
    for(int i = 0; i <= 63; i++) {
      if (Put(m_board, i, turn + 1, NOT_CHANGE)) {
	passp = 0;
	X_SetButtonActive(XW_cell[i], True);
      } else {
	X_SetButtonActive(XW_cell[i], False);
      }
    }
    if (passp) {
      X_message(X_appres.str_humanpass, X_Colorname[turn]);
      X_status = HUMANPASS;
      X_SetButtonActive(XW_pass_btn, True);
    } else {
      X_message(X_appres.str_humanturn, X_Colorname[turn]);
      X_SetButtonActive(XW_pass_btn, False);
    }
    X_start_human_update_timer();
    SetTime();
  }
}

// setup widgets and start game
static void XC_initiategame(Widget, XtPointer, XtPointer)
{
  XtPopdown(XpW_popup);

  player_name[0] = e_list[player_id[0]].name;
  player_name[1] = e_list[player_id[1]].name;

  XtVaSetValues(XW_score_name[0], 
		XtNlabel, player_name[0],
		NULL);
  XtVaSetValues(XW_score_name[1], 
		XtNlabel, player_name[1],
		NULL);

  XtVaSetValues(XW_handslog,
		XtNlabel, "",
		NULL);

  X_SetButtonActive(XW_savelog_btn, False);
  time(&X_starttime);
  strcpy(X_logstring, "");

  initiate_game_1();

  CopyBoard(m_board, v_board, NORM);
  X_update_state();
  XFlush(X_d);

  initiate_game_2();
  
  X_setup_for_next();
}

// Event: enter cell
static void XE_entercell(Widget w, XtPointer cdat, XEvent *, Boolean *cont)
{
  *cont = True;
  int p = (int)cdat;
  if (!XtIsSensitive(w))
    return;

  if (X_status != HUMANTURN)
    return;
  
  char tboard[64];
  CopyBoard(m_board, tboard, NORM);
  if (0 == Put(tboard, p, turn + 1, R_CHANGE))
    return;

  for(int q = 0; q <= 63; q++) {
    if (tboard[q] < 0)
      X_hilight_cell(q, true);
  }
}

// Event: leave cell
static void XE_leavecell(Widget, XtPointer, XEvent *, Boolean *cont)
{
  *cont = true;
  X_unhilight_board();
}

// Callback: cell pushed
static void XC_pushcell(Widget, XtPointer cdat, XtPointer)
{
  int i = (int)cdat;

  if (X_status != HUMANTURN ||
      !Put(m_board, i, turn + 1, NOT_CHANGE)) {
    // shouldn't happen
    XBell(X_d, 50);
    return;
  }
  StopTime();
  X_stop_human_update_timer();
  X_unhilight_board();
  X_put_stone(i);
}

// Callback: Pass button pushed
static void XC_pass(Widget, XtPointer, XtPointer)
{
  if (X_status != HUMANPASS)
    return;

  StopTime();
  X_stop_human_update_timer();
  X_pass_stone();
}

// Callback: Quit button pushed
static void XC_quitapp(Widget, XtPointer, XtPointer)
{
  exit(0);
}

// Callback: Logsave button pushed
static void XC_savelog(Widget, XtPointer, XtPointer)
{
  if (X_status != NOTSTARTED || !X_starttime)
    return;

  if (X_appres.logsavefile[0] == '\0')
    return;

  char *timestr = ctime(&X_starttime);

  FILE *f = fopen(X_appres.logsavefile, "a");

  if (!f) {
    X_message("Log save failed.");
    return;
  }

  fputs(timestr, f);
  fprintf(f, "  %s %s\n  ", player_name[0], player_name[1]);
  fputs(X_logstring, f);
  fputs("\n\n", f);
  fclose(f);

  X_message("Log written to %s.", X_appres.logsavefile);

  X_SetButtonActive(XW_savelog_btn, False);
}

// Message: window manager message
static void XM_destroy_message(Widget, XEvent *e, String *, Cardinal *)
{
  if (e->type == ClientMessage &&
      e->xclient.message_type == XA_wm_protocols &&
      (unsigned)e->xclient.data.l[0] == XA_wm_delete_window) {
    exit(0);
  }
}

// Callback: Popup: OK button pushed
static void XC_selectplayer(Widget, XtPointer, XtPointer)
{
  XtPopup(XpW_popup, XtGrabNone);
}

// Callback: Popup: Cancel button pushed
static void XC_canceldialog(Widget, XtPointer, XtPointer)
{
  XtPopdown(XpW_popup);
}

// Callback: Popup: list entry changed
static void XC_userselected(Widget, XtPointer cdat, XtPointer rdat)
{
  int i = (int)cdat;
  XawListReturnStruct *lrs = (XawListReturnStruct *)rdat;

  assert(i < 2);

  player_id[i] = lrs->list_index;
}

////////////////////////////////////////////////////////////////
// Initialize Routines
////////////////////////////////////////////////////////////////

#define CLASS_NAME "XOthello"

// Configure Data

static String X_fallbacks[] = { // fallback resources
// global
  #include "Xaw3d_colors.inc"

  "*Label.justify: left",
  "*Label.resize: False",
  "*Command.resize: False",
  "*showGrip: off",

// title
  "*game*title.Label: Othello X GUI version " VERSION,
  "*game*title.justify: center",

// status
  "*statusBar.hSpace: 0",
  "*statusBar.pad1.width: 120",
  "*turnLabel.Label: Turn",
  "*turn.justify: right",
  "*turn.width: 30",
  "*turnLabelAfter.Label:",
  "*statusBar.pad2.width: 20",
  "*timeLimitLabel.Label: Time Limit",
  "*timeLimit.justify: right",
  "*timeLimit.width: 50",

// scores
  "*game*scores.white.fromHoriz: black",
  "*game*scores.?.background: gray80",
  "*game*scores.?.resize: false",

  "*game*scores.black.colorLabel.label: Black:",
  "*game*scores.white.colorLabel.label: White:",

  "*game*scores*name.fromHoriz: colorLabel",
  "*game*scores*name.horizDistance: 0",
  "*game*scores*name.width: 115",

  "*game*scores*scoreLabel.label: Discs",
  "*game*scores*scoreLabel.fromVert: colorLabel",
  "*game*scores*score.justify: right",
  "*game*scores*score.width: 23",
  "*game*scores*score.horizDistance: 0",
  "*game*scores*score.fromVert: colorLabel",
  "*game*scores*score.fromHoriz: scoreLabel",

  "*game*scores*timeLabel.label: Time",
  "*game*scores*timeLabel.fromVert: colorLabel",
  "*game*scores*timeLabel.fromHoriz: score",
  "*game*scores*time.width: 50",
  "*game*scores*time.horizDistance: 0",
  "*game*scores*time.fromVert: colorLabel",
  "*game*scores*time.fromHoriz: timeLabel",
  "*game*scores*time.justify: right",

// board
  "*othelloBoard.background: darkgreen",
  "*othelloBoard.width: 340",
  "*othelloBoard.height: 340",
  "*cell.background: darkgreen",
  "*cell.internalWidth: 2",
  "*cell.internalHeight: 2",
  "*cell.foreground: white",

// play bar
  "*pass.label: Pass",

// text region
  "*message.label: Push [Start] to start game.",

// command bar
  "*start.label: Start",
  "*quit.label: Quit",

// popup
  "*select*Label.width: 200",
  "*select*Label.height: 20",
  "*select*black.borderWidth: 0",
  "*select*white.borderWidth: 0",
  "*select*black*label.label: Black Player:",
  "*select*white*label.label: White Player:",
  "*select*white.fromHoriz: black",
  "*select*view.fromVert: label",
  "*select*view.allowVert: true",
  "*select*List.width: 200",
  "*select*Viewport.width: 200",
  "*select*Viewport.height: 200",
  "*select*ok.fromVert: black",
  "*select*ok.label: OK",
  "*select*cancel.label: Cancel",
  "*select*cancel.fromHoriz: ok",
  "*select*cancel.fromVert: black",

  NULL
};

static XrmOptionDescRec opts_parse[] = { // command-line options
  { "-help", ".help", XrmoptionNoArg, (XPointer)"True" },
  { "--help", ".help", XrmoptionNoArg, (XPointer)"True" },
  { "-?", ".help", XrmoptionNoArg, (XPointer)"True" },
  { "-interval", ".interval", XrmoptionSepArg, (XPointer)NULL },
  { "-graycolor", ".grayColor", XrmoptionSepArg, (XPointer)NULL },
  { "-timelimit", ".timeLimit", XrmoptionSepArg, (XPointer)NULL },
  { "-notimelimit", ".timeLimit", XrmoptionNoArg, (XPointer)0 },
  { "-logfile", ".logFile", XrmoptionSepArg, (XPointer)NULL },
  { "-nologfile", ".logFile", XrmoptionNoArg, (XPointer)"" },
};

static XtResource X_resource_table[] = {
  { "help", "Help", XtRBoolean, sizeof(Boolean), 
    XtOffset(X_appres_t *, help), XtRImmediate, (XtPointer)False },
  { "interval", "Interval", XtRMilliInt, sizeof(int),
    XtOffset(X_appres_t *, interval), XtRImmediate, (XtPointer)1000 },
  { "timeLimit", "timeLimit", XtRMilliInt, sizeof(int),
    XtOffset(X_appres_t *, timelimit), XtRImmediate, (XtPointer)0 },
  { "grayColor", "GrayColor", XtRPixel, sizeof(Pixel),
    XtOffset(X_appres_t *, gray_color), XtRString, (XtPointer)"gray80" },

  { "logFile", "LogFile", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, logsavefile), XtRString, (XtPointer)"" },

// String Instance
  { "blackString", "BlackString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_B), XtRString, (XtPointer)"black" },
  { "whiteString", "WhiteString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_W), XtRString, (XtPointer)"white" },
  { "blackCapsString", "BlackString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_Bcaps), XtRString, (XtPointer)"Black" },
  { "whiteCapsString", "WhiteString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_Wcaps), XtRString, (XtPointer)"White" },
  { "handsLogString", "HandsLogString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_handslog), XtRString, 
    (XtPointer)"Turn %d (%s : %s) = %s" },
  { "gameEndString", "GameEndString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_gameend), XtRString, 
    (XtPointer)"Game Finished.  %s (%s) lose. (diff=%d)" },
  { "gameDrawString", "GameDrawString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_gamedraw), XtRString, 
    (XtPointer)"Game Finished.  Draw. (diff=0)." },
  { "gameViolationString", "GameViolationString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_violation), XtRString, 
    (XtPointer)"%s (%s) bad placement (%d). Disqualified." },
  { "compTurnString", "CompTurnString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_compturn), XtRString, (XtPointer)"%s (%s)'s turn." },
  { "humanTurnString", "HumanTurnString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_humanturn), XtRString, 
    (XtPointer)"%s (human)'s turn. Put your disc." },
  { "humanPassString", "HumanTurnString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_humanpass), XtRString, 
    (XtPointer)"%s (human)'s turn. You have to Pass." },
  { "compTimeOverString", "CompTimeOverString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_timeover), XtRString,
    (XtPointer)"%s (%s) time-limit exceeded. Penalty introduced." },
  { "noneString", "NoneString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_none), XtRString, (XtPointer)"None" },
  { "overString", "OverString", XtRString, sizeof(XtRString),
    XtOffset(X_appres_t *, str_over), XtRString, (XtPointer)"Over" },
//  { "", "", XtRString, sizeof(XtRString),
//    XtOffset(X_appres_t *, str_), XtRString, (XtPointer)"" },
};

static String X_shell_translation = // translations of shell window
  "<ClientMessage>: destroy()";

static XtActionsRec X_actions[] = { // actions
  {"destroy", XM_destroy_message },
};

#define XAW3D_NOSHADOW "shadowWidth", 0, XtNborderWidth, 0, 
#define CELLSIZE 36
#define CELLMARGIN 5

// X Application context initialize

static Bool X_initialize_application(int *argcp, char **argv)
{
  XtToolkitInitialize();
  X_app = XtCreateApplicationContext();
  XtSetLanguageProc(X_app, NULL, NULL);

  XtAppSetTypeConverter(X_app, XtRString, XtRMilliInt, X_TypeConvert_MilliInt,
			NULL, 0, XtCacheAll, NULL);

  XtAppSetFallbackResources(X_app, X_fallbacks);
  X_d = XtOpenDisplay(X_app, (String)NULL, NULL, CLASS_NAME, 
		      opts_parse, XtNumber(opts_parse), argcp, argv);
  XW_shell = XtVaAppCreateShell(NULL, CLASS_NAME, applicationShellWidgetClass, X_d, NULL);

  if (!XW_shell) // window not created
    return False;

  X_root = RootWindow(X_d, 0);

  XtVaGetApplicationResources(XW_shell, (XtPointer)&X_appres, X_resource_table,
			      XtNumber(X_resource_table), NULL);

  if (X_appres.interval < 1)
    X_appres.interval = 1;

  X_Colorname[0] = X_appres.str_Bcaps;
  X_Colorname[1] = X_appres.str_Wcaps;
  X_colorname[0] = X_appres.str_B;
  X_colorname[1] = X_appres.str_W;
  time_limit = X_appres.timelimit * 1000;
  if(time_limit <= 0)
    time_limit = INT_MAX;
  
  XtAppAddActions(X_app, X_actions, XtNumber(X_actions));

  return true;
}

static void X_create_selectwindow()
{
  // Setup List
  static char *names[100];
  int nlist;
  for(nlist = 0; e_list[nlist].name; nlist++) {
    names[nlist] = e_list[nlist].name;
  }
  names[nlist] = NULL;

  // widgets
  XpW_popup = XtVaCreatePopupShell("select", transientShellWidgetClass, XW_shell,
				  NULL);

  XpW_frame = XtVaCreateManagedWidget("form", formWidgetClass, XpW_popup, NULL);
  XpW_f0 = XtVaCreateManagedWidget("black", formWidgetClass, XpW_frame, NULL);
  XpW_label0 = XtVaCreateManagedWidget("label", labelWidgetClass, XpW_f0, 
				     NULL);
  XpW_list0 = XtVaCreateManagedWidget("player", listWidgetClass,
				    XtVaCreateManagedWidget("view", viewportWidgetClass, 
							    XpW_f0, NULL),
				    XtNlist, names,
				    NULL);
  XawListHighlight(XpW_list0, 0);
  XtAddCallback(XpW_list0, XtNcallback, XC_userselected, 0);

  XpW_f1 = XtVaCreateManagedWidget("white", formWidgetClass, XpW_frame, NULL);
  XpW_label1 = XtVaCreateManagedWidget("label", labelWidgetClass, XpW_f1,
				     NULL);
  XpW_list1 = XtVaCreateManagedWidget("player", listWidgetClass,
				    XtVaCreateManagedWidget("view", viewportWidgetClass, 
							    XpW_f1, NULL),
				    XtNlist, names,
				    NULL);
  XawListHighlight(XpW_list1, 1);
  XtAddCallback(XpW_list1, XtNcallback, XC_userselected, (XtPointer)1);
  
  XpW_okbtn = XtVaCreateManagedWidget("ok", commandWidgetClass, XpW_frame,
				    NULL);
  XtAddCallback(XpW_okbtn, XtNcallback, XC_initiategame, NULL);

  XpW_cancelbtn = XtVaCreateManagedWidget("cancel", commandWidgetClass, XpW_frame,
				      NULL);
  XtAddCallback(XpW_cancelbtn, XtNcallback, XC_canceldialog, NULL);

  // get atom
  XA_wm_protocols = XInternAtom(X_d, "WM_PROTOCOLS", False);
  XA_wm_delete_window= XInternAtom(X_d, "WM_DELETE_WINDOW", False);
}

#include "pics/empty.xpm"
#include "pics/black.xpm"
#include "pics/black-gray.xpm"
#include "pics/black-plus.xpm"
#include "pics/black-star.xpm"
#include "pics/white.xpm"
#include "pics/white-gray.xpm"
#include "pics/white-plus.xpm"
#include "pics/white-star.xpm"

static void X_create_gamewindow()
{
  // pixmap load
#define LOAD(from,to) XpmCreatePixmapFromData(X_d, X_root, from, to, NULL, NULL)
  LOAD(empty_xpm, &pix_empty);
  LOAD(black_xpm, &pix_black[0]);
  LOAD(black_gray_xpm, &pix_black[1]);
  LOAD(black_plus_xpm, &pix_black[2]);
  LOAD(black_star_xpm, &pix_black[3]);
  LOAD(white_xpm, &pix_white[0]);
  LOAD(white_gray_xpm, &pix_white[1]);
  LOAD(white_plus_xpm, &pix_white[2]);
  LOAD(white_star_xpm, &pix_white[3]);
#undef LOAD

  // master window
  XW_form = XtVaCreateManagedWidget("game", panedWidgetClass, XW_shell, NULL);

  // title bar
  XtVaCreateManagedWidget("title", labelWidgetClass, XW_form, NULL);

  XW_statusbar = XtVaCreateManagedWidget("statusBar", boxWidgetClass, XW_form, NULL);
  XtVaCreateManagedWidget("pad1", coreWidgetClass, XW_statusbar,
			  XtNheight, 1,
			  XAW3D_NOSHADOW
			  NULL);
  XtVaCreateManagedWidget("turnLabel", labelWidgetClass, XW_statusbar,
			  XAW3D_NOSHADOW
			  NULL);
  XW_turn_num = XtVaCreateManagedWidget("turn", labelWidgetClass, XW_statusbar,
					XtNlabel, "0",
					NULL);
  XtVaCreateManagedWidget("turnLabelAfter", labelWidgetClass, XW_statusbar,
			  XAW3D_NOSHADOW
			  NULL);
  XtVaCreateManagedWidget("pad2", coreWidgetClass, XW_statusbar,
			  XtNheight, 1,
			  XAW3D_NOSHADOW
			  NULL);
  char buf[40];
  if (time_limit == INT_MAX) {
    strcpy(buf, X_appres.str_none);
  } else {
    convert_timestring(buf, time_limit, false);
  }
  XtVaCreateManagedWidget("timeLimitLabel", labelWidgetClass, XW_statusbar,
			  XAW3D_NOSHADOW
			  NULL);
  XtVaCreateManagedWidget("timeLimit", labelWidgetClass, XW_statusbar,
			  XtNlabel, buf,
			  NULL);

  // score board
  XW_scores = XtVaCreateManagedWidget("scores", formWidgetClass, XW_form, NULL);

  for(int i = 0; i < 2; i++) {
    static char *widnames[2] = { 
      "black", "white",
    };
    XW_score_board[i] = XtVaCreateManagedWidget(widnames[i], 
						formWidgetClass, XW_scores, NULL);
    XtVaCreateManagedWidget("colorLabel", labelWidgetClass, 
			    XW_score_board[i], 
			    XAW3D_NOSHADOW
			    NULL);
    XW_score_name[i] = XtVaCreateManagedWidget("name", labelWidgetClass, 
					       XW_score_board[i],
					       XtNlabel, "",
					       XAW3D_NOSHADOW
					       NULL);
    XtVaCreateManagedWidget("scoreLabel", labelWidgetClass, 
			    XW_score_board[i], 
			    XAW3D_NOSHADOW
			    NULL);
    XW_score_num[i] = XtVaCreateManagedWidget("score", labelWidgetClass, 
					      XW_score_board[i], 
					      XtNlabel, "0",
					      NULL);
    XtVaCreateManagedWidget("timeLabel", labelWidgetClass, 
			    XW_score_board[i], 
			    XAW3D_NOSHADOW
			    NULL);
    XW_score_time[i] = XtVaCreateManagedWidget("time", labelWidgetClass, 
					       XW_score_board[i], 
					       XtNlabel, "0.0000",
					       NULL);
  }

  // game board
  XW_board = XtVaCreateManagedWidget("othelloBoard", compositeWidgetClass, XW_form, NULL);

  for (int j = 0; j <= 63; j++) {
    int x = j % 8, y = j / 8;
    unsigned xgeom = x * (CELLSIZE + CELLMARGIN) + CELLMARGIN;
    unsigned ygeom = y * (CELLSIZE + CELLMARGIN) + CELLMARGIN;
    XW_cell[j] = XtVaCreateManagedWidget("cell", commandWidgetClass, XW_board, 
					 XtNbitmap, pix_empty,
					 XtNx, xgeom,
					 XtNy, ygeom,
					 NULL
					 );
    XtAddCallback(XW_cell[j], XtNcallback, XC_pushcell, (XtPointer)j);
    XtInsertEventHandler(XW_cell[j], EnterWindowMask,
			 False, XE_entercell, (XtPointer)j, XtListTail);
    XtInsertEventHandler(XW_cell[j], LeaveWindowMask,
			 False, XE_leavecell, (XtPointer)j, XtListTail);
  }

  // game play bar
  XW_nav_bar = XtVaCreateManagedWidget("playBar", boxWidgetClass, XW_form, NULL);
  XW_pass_btn = XtVaCreateManagedWidget("pass", commandWidgetClass, XW_nav_bar, NULL);
  XtAddCallback(XW_pass_btn, XtNcallback, XC_pass, 0);

  // message bar
  XW_handslog = XtVaCreateManagedWidget("lastHands", labelWidgetClass, XW_form, 
					XtNlabel, "", NULL);
  XW_message = XtVaCreateManagedWidget("message", labelWidgetClass, XW_form, NULL);

  // command bar
  XW_command_bar = XtVaCreateManagedWidget("commandBar", boxWidgetClass, XW_form, NULL);
  XW_start_btn = XtVaCreateManagedWidget("start", commandWidgetClass, XW_command_bar, NULL);
  XtAddCallback(XW_start_btn, XtNcallback, XC_selectplayer, 0);
  XW_quit_btn = XtVaCreateManagedWidget("quit", commandWidgetClass, XW_command_bar, NULL);
  XtAddCallback(XW_quit_btn, XtNcallback, XC_quitapp, 0);
  XW_savelog_btn = XtVaCreateManagedWidget("logSave", commandWidgetClass, XW_command_bar, 
					   XtNmappedWhenManaged, X_appres.logsavefile[0] ? True : False,
					   XtNsensitive, False,
					   NULL);
  XtAddCallback(XW_savelog_btn, XtNcallback, XC_savelog, 0);

  // initialize board
  static char dummy_state[] = {
    0,0,0,0,0,0,0,0,
    1,1,1,2,2,0,1,1,
    0,1,2,0,0,1,0,0,
    0,1,0,2,0,1,0,1,
    0,1,0,0,2,1,0,1,
    0,1,2,2,0,0,1,1, // TSG :>
    0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,
  };
  X_setboard(dummy_state, -1);

  X_setstate_startwait();

  XtRealizeWidget(XW_shell);
  
  // get some values
  XtVaGetValues(XW_cell[0],
		XtNbackground, &XCOL_dgreen, 
		NULL);
  
  XCOL_gray = X_appres.gray_color;
  
  // Accept WM Delete
  XtAugmentTranslations(XW_shell, 
			XtParseTranslationTable(X_shell_translation));
  XSetWMProtocols(X_d, XtWindow(XW_shell), &XA_wm_delete_window, 1);
}

static void usage()
{
  char buf[20], *nop;

  if (X_appres.timelimit) {
    sprintf(buf, "%g", X_appres.timelimit /  1000.0);
    nop = "";
  } else {
    strcpy(buf, "");
    nop = "no";
  }

  printf("\
XOthello "VERSION": Weakest Othello Championship  (X-GUI Version)\n\
  othello part (c) 1997 Hideo Kihara (Gel)\n\
  GUI part (c) 1997 Yutaka Oiwa (Aleph-0)\n\
\n\
  usage: XOthello [-help] [-interval <second>] [-graycolor <color>]\n\
                  [-timelimit <second> | -notimelimit]\n\
                  [X toolkit options]\n\
\n\
       -interval <sec>     interval time before computer thinking\n\
       -timelimit <sec>    time limit for computer routine\n\
       -notimelimit        no time limit\n\
       -graycolor <color>  gray color used in interactive mode\n\
\n\
  Current Value: -interval %g -%stimelimit %s\n\
\n\
  Entry List:\n\n", X_appres.interval / 1000.0, nop, buf);
  for(int i = 1; e_list[i].name; i++) {
    printf("    %3d: %s\n", i, e_list[i].name);
  }
  puts("\n");
}

int main(int argc, char **argv)
{
  // for compatibility with othello.o
  assert (e_list[0].func == human);
  e_list[0].func = NULL;

  X_initialize_application(&argc, argv);
  if (X_appres.help || argc != 1) {
    usage();
    exit(0);
  }
  X_create_gamewindow();
  X_create_selectwindow();

  XtAppMainLoop(X_app);
  return 0;
}
